/* Copyright (C) 2008 GFRN systems

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The latest version of this program may be found at
   http://CQiNet.sourceforge.net

   $Log: filters.cpp,v $
   Revision 1.1  2008/04/24 23:54:46  wb6ymh
   Inital import - Basic fixed point BiQuad filters.

*/ 

#include "common.h"
#include "filters.h"

/*
"<S>IIR Filter Specifications -- CHEBYSHEV 1"
"   Type=","LOWPASS"
"   Order= ",4
"   Sampling frequency=",8000
"   Passband:",200,0
"   Stopband:",780.6665,0
"   Passband ripple:",1
"   Stopband loss:",60
"<C>Coefficients (A,D,E,B,C) for wordlength ",24
1,1.653416E-03,2,1,-1.933499,.9572726
2,5.256554E-03,2,1,-1.892893,.8994564
*/

int PlDecoee1Coef[] = {
   27,      // A 1.653416E-03
   -31678,  // B -1.933499
   15684,   // C .9572726
   32768,   // D 2
   16384,   // E 1
};

int PlDecoee2Coef[] = {
   86,      // A 5.256554E-03
   -31013,  // B -1.892893
   14737,   // C .8994564
   32768,   // D 2
   16384,   // E 1
};


/*
 BiQuad, design coefficients 
 Wn = A1 * Xn - B * Wn-1 - C * Wn-2
 Yn = A2 * Wn + (A2*D) * Wn-1 + (A2*E) * Wn-2
 
 where A = A1 * A2
 
 if a2 = 1  then
 
 Wn = A * Xn - B * Wn-1 - C * Wn-2
 Yn = Wn + D * Wn-1 + E * Wn-2

 update
 
 Wn-2 = Wn-1
 Wn-1 = Wn
*/

#define A   Coefficients[0]
#define B   Coefficients[1]
#define C   Coefficients[2]
#define D   Coefficients[3]
#define E   Coefficients[4]
int16 BiQuad(int16 Sample,int *Coefficients,int *W)
{
   int Wn;
   int Yn;

   Wn = (A * Sample) - (B * W[0]) - (C * W[1]);
   Yn = Wn + (D * W[0]) + (E * W[1]);

   W[1] = W[0];
   W[0] = (Wn >> 14);

   return (Yn >> 14);
}
#undef A
#undef B
#undef C
#undef D
#undef E

/*
   Wn = A1 * Xn - C * Wn-1
   Yn = A2 * Wn + (A2 * E) * Wn-1
   
   Wn-1 = Wn
   
   where A = A1 * A2
 
   if a2 = 1  then
   
   Wn = A * Xn - C * Wn-1
   Yn = Wn + E * Wn-1
   
*/ 
#define A   Coefficients[0]
#define C   Coefficients[1]
#define E   Coefficients[2]
int16 FirstOrder(int16 Sample,int *Coefficients,int *W)
{
   int Wn;
   int Yn;

   Wn = (A * Sample) - (C * W[0]);
   Yn = Wn + (E * W[0]);

   W[0] = (Wn >> 14);

   return (Yn >> 14);
}
#undef  A
#undef  C
#undef  E

void CtcssLowPass(int16 *In,int16 *Out,int Samples,int *Temp)
{
   int i;
   int16 tmp;

   for(i = 0; i < Samples; i++) {
      tmp = BiQuad(In[i],PlDecoee1Coef,Temp);
      Out[i] = BiQuad(tmp,PlDecoee2Coef,&Temp[2]);
   }
}


