/* Copyright (C) 2002 GFRN systems

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The latest version of this program may be found at
   http://CQiNet.sourceforge.net

   $Log: nodes.h,v $
   Revision 1.4  2008/04/25 05:13:57  wb6ymh
   Added  support for site prefix decoding.

   Revision 1.3  2008/03/18 20:04:14  wb6ymh
   Added bCwIsID and bAnnoucementIsID.

   Revision 1.2  2008/03/11 14:08:45  wb6ymh
   1. Moved NetAudio here from conference.h.
   2. Deleted LNK_FLG_AUDIO_RDY from LinkedNode.
   3. Added BufNum to LinkedNode.
   4. Added BufNum to Node class.

   Revision 1.1  2008/03/09 17:10:13  wb6ymh
   Inital import - files broken out of old audio.cpp and audio.h due to
   major restructuring of the internal infrastructure to treat VoIP connections
   as pseudo ports.

*/

#ifndef _NODE_H_
#define _NODE_H_
#ifdef USB_SUPPORT
#include <usb.h>
#endif
#include "ToneGen.h"
#include "sf.h"
#include "adpcm.h"
#include "gsm.h"
#include "ilink.h"
#include "tone_detect.h"
#include "super_tone_rx.h"
#include "tone_generate.h"
#include "dtmf.h"
#define  AUDIO_BUF_SIZE    ILINK_AUDIO_SAMPLES_PER_PACKET

typedef struct PlayPCM_TAG {
   struct PlayPCM_TAG *Link;  // NB: must be first
   int   Flags;
      #define PLAY_FLAG_ID       0x1   // voice ID
      #define PLAY_FLAG_TTS      0x2   // FileName = Text for TTS system
      #define PLAY_FLAG_TTS_OUT  0x4   // audio rendered by TTS system
      #define PLAY_FLAG_NO_PAUSE 0x7   // Types that can't be paused
      
   char  *FileName;
   int ReplyPort; 
} PlayPCM;

typedef struct DTMF_Command_Tag {
   struct DTMF_Command_Tag *Link;   // NB: must be first
   char  *DTMF_String;
   char  *Command;
} DTMF_Command;

typedef struct NetAudioTAG {
   struct NetAudioTAG *Link;  // NB: must be first
   int   Available;     // bytes of available data (samples *2)
   int   Users;         // number of nodes yet to consume this buffer
   int   BufNum;        // sequencial buffer number
   char  Buf[1];        // actual length varies
} NetAudio;

typedef struct LinkedNodeTAG {
   struct LinkedNodeTAG *pNextTx; // For the receiver nodes use
   struct LinkedNodeTAG *pNextRx; // For the transmitter nodes use
   int Flags;
      #define LNK_FLG_CTONE         1  // Send courtesy to this tx
      #define LNK_FLG_CTCSS_FILTER  2  // filter ctcss before sending to this tx
      #define LNK_FLG_DEEMPHASIS    4  // receiver wants deemphasized audio
   class Node *pRx;
   class Node *pTx;

   NetAudio *pInHead;   // used by VoipNode to queue pcm
   int   Used;          // bytes of data used from current NetAudio buffer
   int   BufNum;        // Last buffer consumed
} LinkedNode;

class NamedNode {
public:
   virtual ~NamedNode();   // NB: must be here for a kludge casts to work!
   char *NodeName;
};

class Node : public NamedNode {
public:
// public functions   
   Node();
   virtual ~Node();
   void SaveDTMF(char Digit);
   void DecodeDTMF(char *Command);
   int CancelToneGen(int id);
   void AddToneGen(CToneGen *pGenerator);
   virtual int WriteAudioOut(ClientInfo *p) = 0;
   int LoadCommandFile(struct config_entry *p, char *Arg);
   void SayPhrase(char *Phrase,int ReplyPort = 0);
   void SayCall(char *Call);
   virtual int GetRxAudio(LinkedNode *pLink,int16 **pAudio) = 0;
   virtual void AddTx(LinkedNode *pLink);
   virtual int DeleteTx(LinkedNode *pLink);
   virtual void AddRx(Node *pNode);
   virtual int DeleteRx(Node *pNode);
   int UnlinkAll();
   int Shutup();
   void UpdateCtoneFlags();
   int InitToneBuf(struct config_entry *p, char *Arg);
   virtual int Enabled();

// Config variables
   int DtmfCTCSSMode;
   int Min1stDigitHoldTime;
   int MuteDTMF;
   int DtmfCoverTone;
   int MinDtmfInterdigitTime;
   int DtmfMethod;
   int Twist;
   int ReverseTwist;
   int Ctone;
   int CtoneDelay;
   int TxTimeout;
   int RxTimeout;
   int TxTimeoutTone;
   int RxTimeoutTone;
   int CtoneResetsTimeout;
   int CTCSSLevel;
   int CWToneLevel;
   int CWSpeed;
   int CWAutoJam;
   int CWToneFreq;
   int PrefixedTone;
   int DTMFCommandTimeout;
   int PrefixTimeout;

   float EchoCTCSSFreq;

   char *Dtmf1stDigits;
   char *CommandFilename;
   char *CtoneMap;
   char *DTMFPrefix;

// Variables accessed by C, eventually to be moved to protected ?
   dtmf_rx_state_t DtmfRxState;
   LinkedNode *pReceivers; // Linked list of receivers feeding this transmitter
   LinkedNode *pTransmitters; // Linked list of transmitters this receiver feeds
   int Transmitters;    // number of transmitters this receiver feeds

protected:
   DTMF_Command *pDTMFCommands;
   struct timeval LastDTMFDigit;
   int bRxActive;
   int bCtcssActive;
   #undef MAX_DTMF_DIGITS
   #define MAX_DTMF_DIGITS    16
   char  DTMF_Buf[MAX_DTMF_DIGITS+1];
   int   DTMF_Digits;
   CToneGen *ToneGenerators;
   FILE *Playfp;
   PlayPCM *pPlayHead;
   PlayPCM *pPlayTail;
   int b8BitFile;
   time_t LastActivity;
   int PlayAvail;
   int MaxMixedLevel;
   int LastMixGain;
   int16 AudioBuf[AUDIO_BUF_SIZE];
   unsigned int AudioAvailable;    // number of BYTES of audio in AudioBuf
   char *CWChar;
   int bTxTimeOut;
   CToneGen *TxTimeoutToneGen;
   double CTCSSAngle;
   int ElementCount;
   int bCwToneOn;
   int bElementSpace;
   int bRxTimeout;
   int RampCount;
   char *Element;
   time_t TxKeyTime;
   struct timeval CtoneTimer;
   int BufNum;
   int bCwIsID:1;          // current CW message is an transmitter ID
   int bAnnoucementIsID:1; // current annoucement is an transmitter ID
   int bPrefixed:1;        // site prefix decoded

   int PlayStoredSound(int ID);
   int GetAnnounceAudio();
   void StartNextAnnoucement();
   void FreeAnnoucement();
   void PlayAnnouncementFile(char *FileName,int Flags,int ReplyPort = 0);
   void ScaleMixedAudio();
   void MixAudio(int16 *In,int Samples);
   void GetTxAudio();
   void GenCW(int16 *CwBuf);
   void GenCTCSS(int16 *CTCSSBuf);
   void SendCtone(int bDelayed);
   void CheckDtmfTimeouts(void);
   int CheckPrefix(void);

private:
   int DitSamples;
   int CwRampCount;     // number of samples ramping up and down from full level
};


extern int bTTS_Running;

int InitToneBuf(struct config_entry *p, char *Arg, AfCmd Cmd, int x);
int NodeCompare(const void *avl_a,const void *avl_b,void *avl_param);
char *FindToneSpec(int ID);
Node *FindNode(char *NodeName);
void DtmfCallback(void *user_data, int code, int level);

#endif   //_NODE_H_

