/* Copyright (C) 2007 GFRN systems

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The latest version of this program may be found at
   http://CQiNet.sourceforge.net

   $Log: ports.cpp,v $
   Revision 1.5  2008/04/25 05:23:23  wb6ymh
   1. Began modifications to allow a single stereo sound device to be use to
      interface to two separate radios.
   2. Modified the code to use 48khz to talk CM108 based USB dongles.  Added
      local sampling rate conversion routines to convert between 8Khz and 48Khz.
   3. Added support for site prefix decoding.
   4. Removed unneeded ClientInfo pointer arguments to ProcessAudioIn, PollCOS,
      and PollDTMF.
   5. Completed support for software based CTCSS decoding including search
      and decode all tones options.
   6. Added SetRxTone function.

   Revision 1.4  2008/03/18 20:05:19  wb6ymh
   Added code to set bCWIsID before sending CW id.

   Revision 1.3  2008/03/11 14:11:30  wb6ymh
   1. Moved include for common.h to after system headers.  Fixes compile problems
      (min macro) on some systems.
   2. Modified GetRxAudio to use BufNum to determine if new audio is ready rather
      than LNK_FLG_AUDIO_RDY. Just an optimization and code cleanup.

   Revision 1.2  2008/03/09 18:19:47  wb6ymh
   Added headers needed by FreeBSD.

   Revision 1.1  2008/03/09 17:10:13  wb6ymh
   Inital import - files broken out of old audio.cpp and audio.h due to
   major restructuring of the internal infrastructure to treat VoIP connections
   as pseudo ports.

*/ 
// FreeBSD, Linux, etc..
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#ifdef TIME_WITH_SYS_TIME
   #include <sys/time.h>
   #include <time.h>
#else
   #ifdef HAVE_SYS_TIME_H
      #include <sys/time.h>
   #else
      #include <time.h>
   #endif
#endif
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/soundcard.h>
#ifdef HAVE_FCNTL_H
   #include <fcntl.h>
#endif
#include <netdb.h>
#include <ctype.h>
#ifdef HAVE_SYS_TIMEB_H
   #include <sys/timeb.h>
#endif
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <math.h>
#include <typeinfo>
#include <sys/socket.h>
#include <netinet/in.h>


#ifdef HAVE_SYS_SYSLIMITS_H
#include <sys/syslimits.h> // for PATH_MAX
#endif
#include <sys/file.h>

#include "common.h"
#include "avl.h"
#include "main.h"
#include "configvars.h"
#include "configfile.h"
#include "dirclient.h"
#include "conference.h"
#include "linkbox.h"
#include "eventhook.h"
#include "ctcss.h"
#include "tbd.h"

#include "nodes.h"
#include "ports.h"
#include "voip.h"
#include "filters.h"
#include "usbdev.h"

#ifdef USE_DMALLOC
#include "dmalloc.h"
#endif


// IRLP interface board equates
#define IRLP_LOCKFILE   "/tmp/irlp-lockfile-parport0"


#define IRLP_PTT        0x02
#define IRLP_AUX1       0x04
#define IRLP_AUX2       0x08

#define IRLP_DTMF_SHIFT 3
#define IRLP_COS        0x80  // Busy

#ifdef __linux__
#include <linux/ppdev.h>
#include <linux/parport.h>
#define PP_RD_DATA   PPRDATA
#define PP_WR_DATA   PPWDATA
#define PP_RD_STATUS PPRSTATUS
#define PP_DATA_TYPE char
#define IRLP_DEVICE     "/dev/parport0"
#endif

#ifdef __FreeBSD__
#include <dev/ppbus/ppi.h>
#include <dev/ppbus/ppbconf.h>
#define PP_RD_DATA   PPIGDATA
#define PP_WR_DATA   PPISDATA
#define PP_RD_STATUS PPIGSTATUS
#define PP_DATA_TYPE int
#define IRLP_DEVICE     "/dev/ppi0"
#endif

#define SERIAL_OPEN_FLAGS O_FSYNC | O_APPEND | O_NONBLOCK | O_RDWR | \
        O_NOCTTY | O_NDELAY

#ifndef  DEBUG_PCM_IN
#define AUDIO_OPEN_FLAGS   O_RDWR | O_NONBLOCK
#else
#define AUDIO_OPEN_FLAGS   O_RDWR
#endif

#define RX_IDLE_TOSS_COUNT    2

// global bit map of ports with active signals 
int ActiveRxMap = 0;
int WriteWaits = 0;
int WriteWakeups = 0;
int WriteStalls = 0;

int IRLPfd = -1;
int LockIRLPfd = -1;
int gIRLP_PortLocked = FALSE;

FILE *fp = NULL;
FILE *fp1 = NULL;


struct BaudRateTbl_TAG {
   int   Baudrate;
   int   SpeedConstant;
} BaudRateTbl[] = {
   {110,B110},
   {300,B300},
   {600,B600},
   {1200,B1200},
   {2400,B2400},
   {4800,B4800},
   {9600,B9600},
   {19200,B19200},
   {0}
};

CPort *pMasterPort = NULL;

int16 SilenceBuf[AUDIO_BUF_SIZE];

/*
"<S>IIR Filter Specifications -- BUTTERWORTH"
"   Type=","LOWPASS"
"   Order= ",4
"   Sampling frequency=",48000
"   Passband:",3000,0
"   Stopband:",4000,0
"   Passband ripple:",.5
"   Stopband loss:",3.660441
"<C>Coefficients (A,D,E,B,C) for wordlength ",24
1,3.742148E-02,2,1,-1.475217,.6869044
2,6.127672E-02,2,1,-1.207816,.3811328
*/

int DownConvertLP1Coef[] = {
   613,     // A 3.742148E-02
   -24170,  // B -1.475217
   11254,   // C .6869044
   32768,   // D 2
   16384,   // E 1
};

int DownConvertLP2Coef[] = {
   1004,    // A 6.127672E-02
   -19789,  // B -1.207816
   6244,    // C .3811328
   32768,   // D 2
   16384,   // E 1
};

void SendCWMsg(char *Msg);
void ID();
void PlayAnnouncementFile(char *FileName,int Flags);
void ExecuteCommand(char *Command);
void SetTimeout(struct timeval *pTimer,int Timeout);

int ClaimIRLP() 
{
   int Ret = 0;   // Assume the best
   if(ShareIRLP_Port) {
      if(flock(LockIRLPfd,LOCK_EX) != 0) {
         Ret = errno;
         LOG_ERROR(("%s: flock(LOCK_EX) failed: %s",__FUNCTION__,
                    Err2String(errno)));
      }
#ifdef __linux__
      else if(ioctl(IRLPfd,PPCLAIM) != 0) {
         Ret = errno;
         LOG_ERROR(("%s: ioctl(PPCLAIM) failed: %s",__FUNCTION__,
                    Err2String(errno)));
         if(flock(LockIRLPfd,LOCK_UN) != 0) {
            LOG_ERROR(("%s: flock(LOCK_UN) failed: %s",__FUNCTION__,
                       Err2String(errno)));
         }
      }
#endif
      else {
         gIRLP_PortLocked = TRUE;
      }
   }

   return Ret;
}

void ReleaseIRLP()
{
   if(ShareIRLP_Port && gIRLP_PortLocked) {
#ifdef __linux__
      if(ioctl(IRLPfd,PPRELEASE) != 0) {
         LOG_ERROR(("%s: ioctl(PPRELEASE) failed: %s",__FUNCTION__,
                    Err2String(errno)));
      } else 
#endif
      if(flock(LockIRLPfd,LOCK_UN) != 0) {
         LOG_ERROR(("%s: flock(LOCK_UN) failed: %s",__FUNCTION__,
                    Err2String(errno)));
      }
      else {
         gIRLP_PortLocked = FALSE;
      }
   }
}

CPort::CPort()
{
// set default configuration variable values
   VoxThreshold = 0;
   VoxHoldTime = 0;
   VoxTripDelay = 0;
   TxKeyMethod = 0;
   RxCosMethod = 0;
   SerialBaudRate = 2400;
   DTMFPollTime = 50;
   InvertPTT = 0;
   InvertCOS = 0;
   IdInterval = 0;
   VoiceIDIdleTime = 0;
   HalfDuplexRF = 0;
   HalfDuplex = 0;
   PCMRate = 8000;
// set CtcssDecoderSamples to 300 milliseconds decode time @ 8 Khz sample rate.
   CtcssDecoderSamples = 3200;  
   bMonitor = 0;
   COS_KeyCode = 0;
   CTCSS_KeyCode = 0;

   AudioIn8Bit = 0;
   CWToneFreq = 1000;


   SerialDevName = NULL;
   CWId = NULL;
   VoiceID = NULL;
   AudioDevice = NULL;
   AudioDongleSN = NULL;

// Init class variables
   AudioInAvailable = 0;
   AudioInBuf8 = NULL;
   AudioInBuf16 = NULL;
   AudioInBuf = NULL;
   bEvenRxBuf = 1;
   bDevInCOS = FALSE;
   bCosActive = FALSE;
   bDiscFeed = 0;
   bUngatedRx = 0;
   TossRxIn = 0;
   PTT_Bit = -1;
   COS_Bit = -1;
   CTCSS_Bit = -1;
   pSerialSlave = NULL;
   pParallelSlave = NULL;
   SerialFd = -1;
   HangTimer.tv_sec = 0;
   AudioUsed = 0;
   RxCtcssMethod = 0;
   RxCTCSSMode = 0;
   InvertCTCSS = 0;
   DtmfCTCSSMode = 1;   // require CTCSS if CTCSS is enabled
   pAudioC = NULL;
   AudioChannel = 0;    // default to mono
   pAudioSlave = NULL;
   StereoInBuf = NULL;
   StereoOutBuf = NULL;
   RxCTCSSTone = -1;    // default to carrier squelch
   UCLastSample = 0;
   memset(DCTemp,0,sizeof(DCTemp));
   memset(UCTemp,0,sizeof(UCTemp));
   CtcssDecoderState = NULL;
   RxToneDesc = strdup("disabled");
   bRxCtcssSearch = FALSE;
   bRxCtcssAny = FALSE;
   CtcssRxThreshold = 6000;
   bConvertSampleRate = 0;
   DisableRateConversion = 0;
#ifdef USB_SUPPORT
   udev = NULL;
   UsbDevType = 0;
#endif
};

CPort::~CPort()
{
   if(StereoInBuf != NULL) {
      free(StereoInBuf);
   }

   if(StereoOutBuf != NULL) {
      free(StereoOutBuf);
   }

   if(RxToneDesc != NULL) {
      free(RxToneDesc);
   }

   if(CtcssDecoderState != NULL)  {
      CtcssDestroy(CtcssDecoderState);
   }
}

void CPort::KeyTx(int bKey)
{
   int ModemCtrl;
   PP_DATA_TYPE IRLP_Bits;

   bTxKeyed = bKey;

   if(bKey) {
      D2PRINTF(("K"));
      TxKeyTime = TimeNow.tv_sec;
      if((TimeNow.tv_sec - LastID) > IdInterval) {
      // Time to ID
         ID();
      }
   }
   else {
      D2PRINTF(("U"));
   }

   switch(TxKeyMethod) {
      case 1:  // WB2REM / VA3TO interface
      // For DtmfMethod 1, tx keying is handled in PollDTMF()
         if(DtmfMethod != 1) {
            if(bTxIsKeyed != bTxKeyed) {
               int wrote;
               bTxIsKeyed = bTxKeyed;
               wrote = write(SerialFd,bTxKeyed ? "T" : "R",1);
               if(wrote != 1) {
                  LOG_ERROR(("KeyTx(): write failed: %s",Err2String(errno)));
               }
            }
         }
         break;

      case 2:  // Parallel port interface
         if(ClaimIRLP()) {
            LOG_ERROR(("%s: ClaimIRLP failed\n",__FUNCTION__));
         }
         else if(ioctl(IRLPfd,PP_RD_DATA,&IRLP_Bits) != 0) {
            LOG_ERROR(("%s: ioctl(PP_RD_DATA) failed: %s",__FUNCTION__,
                       Err2String(errno)));
         }
         else {
            if((bKey && !InvertPTT) || (!bKey && InvertPTT)) {
               IRLP_Bits |= PTT_Bit;
            }
            else {
               IRLP_Bits &= ~PTT_Bit;
            }

            if(ioctl(IRLPfd,PP_WR_DATA,&IRLP_Bits)) {
               LOG_ERROR(("%s: ioctl(PP_WR_DATA) failed: %s",__FUNCTION__,
                          Err2String(errno)));
            } 
         }
         ReleaseIRLP();
         break;

      case 3:  // RTS
      case 4:  // DTR
         if(SerialFd != -1) {
            if(ioctl(SerialFd,TIOCMGET,&ModemCtrl) == -1) {
               LOG_ERROR(("KeyTx(): ioctl(TIOCMGET) failed: %s",
                          Err2String(errno)));
            }
            else {
               if((bKey && !InvertPTT) || (!bKey && InvertPTT)) {
                  ModemCtrl |= PTT_Bit;
               }
               else {
                  ModemCtrl &= ~PTT_Bit;
               }
               if(ioctl(SerialFd,TIOCMSET,&ModemCtrl) == -1) {
                  LOG_ERROR(("KeyTx(): ioctl(TIOCMSET) failed: %s",
                             Err2String(errno)));
               }
            }
         }
         break;

      case 5:  // USB GPIO
#ifdef USB_SUPPORT
         UsbKeyRadio(bKey);
#endif
         break;
   }
}

void CPort::ID()
{
   int TimeSinceLastActvity = TimeNow.tv_sec - LastActivity;

   if(!bTxKeyed && HalfDuplexRF && 
      (bTalking || TimeLapse(&LastSpoke) < 5000)) 
   {  // We're running on a half duplex link and we're talking or have been
      // talking recently.  Don't key the transmitter !
   }
   else {
      if(CWAutoJam || TimeSinceLastActvity < IdInterval) {
      // Send an ID
         LastID = TimeNow.tv_sec;

         if(VoiceID != NULL && TimeSinceLastActvity > VoiceIDIdleTime) {
         // We have a voice ID and the channel has been idle long enough 
         // to play it
            PlayAnnouncementFile(VoiceID,PLAY_FLAG_ID);
         }
         else if(CWId != NULL) {
            bCwIsID = TRUE;
            SendCWMsg(CWId);
         }
      }
      else {
      // Don't ID again until we have some new activity
         LastID = 0;
      }
   }
}

int CPort::GetRxAudio(LinkedNode *pLink,int16 **AudioOut)
{
   int Ret = 0;

   if(pLink->BufNum < BufNum) {
      pLink->BufNum = BufNum;
      Ret = AUDIO_BUF_SIZE;
      *AudioOut = AudioInBuf;
   }
   return Ret;
}

// Convert sampling rate of 8Khz data in AudioBuf to 48Khz data in AudioBuf48Khz
void CPort::UpConvert()
{
   int i;
   int j;
   int k = 0;
   int16 Sample = UCLastSample;
   int16 Delta;
   int16 Temp;

   for(i = 0; i < AUDIO_BUF_SIZE; i++) {
      Delta = (AudioBuf[i] - Sample) / 6;
      for(j = 0; j < 6; j++) {
         Sample += Delta;
         Temp = BiQuad(Sample,DownConvertLP1Coef,UCTemp);
         AudioBuf48Khz[k++] = BiQuad(Temp,DownConvertLP2Coef,&UCTemp[2]);
      }
      Sample = AudioBuf[i];
   }
   UCLastSample = Sample;
   AudioAvailable = k * sizeof(int16);
}

// Convert sampling rate of 48Khz data in AudioInBuf48 to 8Khz data in AudioInBuf
void CPort::DownConvert()
{
   int i;
   int j;
   int k = 0;
   int16 Temp;

   for(i = 0; i < AUDIO_BUF_SIZE; i++) {
      Temp = BiQuad(AudioInBuf48Khz[k++],DownConvertLP1Coef,DCTemp);
      AudioInBuf[i] = BiQuad(Temp,DownConvertLP2Coef,&DCTemp[2]);
      for(j = 1; j < 6; j++) {
         Temp = BiQuad(AudioInBuf48Khz[k++],DownConvertLP1Coef,DCTemp);
         BiQuad(Temp,DownConvertLP2Coef,&DCTemp[2]);
      }
   }
}


// return TRUE if we write audio to the sound device
int CPort::WriteAudioOut(ClientInfo *p)
{
   int wrote;
   char *pWriteData;
   int Ret = FALSE;

   bAudioOutCalled = TRUE;
   
   for(; ; ) {
      if(AudioAvailable == 0) {
         AudioUsed = 0;
         D2PRINTF(("F"));
         GetTxAudio();
         if(pAudioSlave != NULL) {
            pAudioSlave->GetTxAudio();
            if(AudioAvailable == 0) {
               AudioAvailable = pAudioSlave->AudioAvailable;
            }
         }
         
         if(AudioChannel != 0 && AudioAvailable != 0) {
            AudioAvailable = AUDIO_BUF_SIZE * sizeof(int16) * 2;
         // pack data from the two mono buffers into the stereo buffer
            int16 *pRight;
            int16 *pLeft;
            int16 *pOut = StereoOutBuf;

            if(AudioChannel == 1) {
               pRight = AudioBuf;
               pLeft = pAudioSlave == NULL ? SilenceBuf : pAudioSlave->AudioBuf;
            }
            else {
               pRight = pAudioSlave == NULL ? SilenceBuf : pAudioSlave->AudioBuf;
               pLeft = AudioBuf;
            }

            for(int i = 0; i < AUDIO_BUF_SIZE; i++) {
               *pOut++ = *pLeft++;
               *pOut++ = *pRight++;
            }
         }

         if(AudioAvailable != 0 && bConvertSampleRate) {
         // Upconvert from 8Khz to 48Khz sampling rate
            UpConvert();
         }
      }

      if(AudioAvailable > 0) {
         if(bConvertSampleRate) {
            pWriteData = ((char *) AudioBuf48Khz) + AudioUsed;
         }
         else if(AudioChannel == 0) {
            pWriteData = ((char *) AudioBuf) + AudioUsed;
         }
         else {
            pWriteData = ((char *) StereoOutBuf) + AudioUsed;
         }

         wrote = write(p->Socket,pWriteData,AudioAvailable);
#if 0
         if(fp != NULL && wrote > 0) {
            fwrite(pWriteData,wrote,1,fp);
         }
#endif
         if(wrote <= 0) {
            if(errno == EAGAIN) {
               WriteWaits++;
               p->bWriteWait = TRUE;
               D2PRINTF(("W"));
            }
            else {
               LOG_ERROR(("WriteAudioOut(): write() failed: %s",
                          Err2String(errno)));
            }
            break;
         }
         else {
            D2PRINTF(("w"));
            HangTimer.tv_sec = 0;
            if(pAudioSlave != NULL) {
               pAudioSlave->HangTimer.tv_sec = 0;
            }
            AudioUsed += wrote;
            Ret = TRUE;
            if(wrote < (int) AudioAvailable) {
            // The audio buffers filled, take a break
               AudioAvailable -= wrote;
               p->bWriteWait = TRUE;
               break;
            }
            AudioAvailable -= wrote;
         }
      }
      else {
      // No more audio available now
         WriteStalls++;
         p->bWriteWait = FALSE;
         break;
      }
   }
   DFLUSH();

   if(Ret) {
      if(!bTxKeyed) {
         KeyTx(TRUE);
      }
      if(pAudioSlave != NULL && !pAudioSlave->bTalking) {
         pAudioSlave->KeyTx(TRUE);
      }
   }

   return Ret;
}

void CPort::StartTalking(int bTalk)
{
   if(bTalking != bTalk) {
      bTalking = bTalk;
      if(!bTalk && DTMF_Digits > 0) {
      // User unkeyed and there's something in the DTMF buffer
         if(MuteDTMF && DtmfCoverTone != 0) {
            CancelToneGen(DtmfCoverTone);
         }
         if(DTMFPrefix == NULL || bPrefixed) {
            DecodeDTMF(DTMF_Buf);
         }
         DTMF_Digits = 0;
      }

      if(!bTalk) {
         LastSpoke = TimeNow;
      }
   }
}

void CPort::Vox()
{
   int AveLevel = 0;
   int i;

   for(i = 0; i < AUDIO_BUF_SIZE; i++) {
      AveLevel += abs(AudioInBuf[i]);
   }

   AveLevel /= AUDIO_BUF_SIZE;
   
   if(AveLevel > VoxThreshold) {
      if(!bTalking) {
         StartTalking(TRUE);
         D2PRINTF(("V"));
         DFLUSH();
      }
      LastVoxTrip = TimeNow;
   }
   else if(bTalking && TimeLapse(&LastVoxTrip) > VoxHoldTime) {
      StartTalking(FALSE);
      D2PRINTF(("v"));
      DFLUSH();
   }
}

void CPort::ProcessAudioIn()
{

   if(HalfDuplexRF && bTxKeyed) {
      StartTalking(FALSE);
   }
   else {
      if(RxCosMethod == 1) {
         Vox();
      }
   }
   
   if(CtcssDecoderState != NULL) {
      int16 CtcssBuf[AUDIO_BUF_SIZE];
      int RxCTCSS;
   
   // low pass filter input
      CtcssLowPass(AudioInBuf,CtcssBuf,AUDIO_BUF_SIZE,CtcssTemp);
      RxCTCSS = UpdateCTCSS(CtcssBuf,AUDIO_BUF_SIZE,CtcssDecoderState,
                            CtcssRxThreshold);
      if(CurrentRxCTCSS != RxCTCSS) {
      // CTCSS decode state has changed
         CurrentRxCTCSS = RxCTCSS;
         if(bRxCtcssAny || bRxCtcssSearch) {
            if(RxToneDesc != NULL) {
               free(RxToneDesc);
            }

            if(RxCTCSS != -1)  {
               RxToneDesc = strdup(freq_ctcss[RxCTCSS]);
            }
            else {
               RxToneDesc = strdup(bRxCtcssAny ? "any" : "search");
            }
         }

         if(bCosActive) {
            if(bRxCtcssAny || bRxCtcssSearch || DebugLevel > 0) {
               if(RxCTCSS != -1)  {
                  LOG_ERROR(("%s: Node %s CTCSS %s\n",__FUNCTION__,
                             NodeName,freq_ctcss[RxCTCSS]));
               }
               else {
                  LOG_ERROR(("%s: Node %s CTCSS decode inactive\n",__FUNCTION__,
                             NodeName));
               }
            }
            if(bRxCtcssAny || bRxCtcssSearch) {
               if(RxCTCSS != -1 && bRxCtcssAny) {
                  EventHook("ctcss_decode %s %s",NodeName,RxToneDesc);
               }
            }
         }
         CtcssDecodeTime = TimeNow;
      }

      if(bRxCtcssSearch && RxCTCSS != -1) {
         if(TimeLapse(&CtcssDecodeTime) > 1000) {
         // Found tone!
            SetSoftRxTone(RxCTCSS);
            EventHook("ctcss_decode %s %s",NodeName,RxToneDesc);

            LOG_ERROR(("Tone search complete, port %s tone %s\n",
                       NodeName,RxToneDesc));
         }
      }
   }

   if(bTalking) {
#if 0
      if(fp != NULL) {
         fwrite(AudioInBuf,AUDIO_BUF_SIZE*2,1,fp);
      }
#endif

      if(DtmfMethod == 3) {
         dtmf_rx(&DtmfRxState,AudioInBuf,AUDIO_BUF_SIZE);
      }
   }

   CheckDtmfTimeouts();

   if(!bTalking && bUngatedRx && TossRxIn < RX_IDLE_TOSS_COUNT) {
   // Unsquelched audio source, toss the first audio buffer after COS goes
   // active to eliminate possible squelch tail on keyup
      D2PRINTF(("I"));
      TossRxIn++;
   }
   else if(bTalking && TossRxIn == RX_IDLE_TOSS_COUNT) {
      TossRxIn = 0;
      D2PRINTF(("i"));
   }
   else if(bTalking && !bMonitor && (!MuteDTMF || DTMF_Digits == 0)) {
      BufNum++;
      D2PRINTF(("r"));
   }
}

void CPort::GetAudioIn(ClientInfo *p)
{
   int Read;
   int FullBufLen;
   char *pWriteData;
   int i;

#ifndef  DEBUG_PCM_IN
   for( ; ; ) {
      if(bConvertSampleRate) {
      // Read a buffer's worth of 48Khz mono data
         pWriteData = ((char *) AudioInBuf48Khz) + AudioInAvailable;
         FullBufLen = AUDIO_BUF_SIZE * sizeof(int16) * 6;
      }
      else if(AudioChannel != 0) {
      // Read a buffer's worth of stereo data
         pWriteData = ((char *) StereoInBuf) + AudioInAvailable;
         FullBufLen = AUDIO_BUF_SIZE * sizeof(int16) * 2;
      }
      else if(AudioIn8Bit) {
         pWriteData = ((char *) AudioInBuf8) + AudioInAvailable;
         FullBufLen = AUDIO_BUF_SIZE;
      }
      else {
         pWriteData = ((char *) AudioInBuf16) + AudioInAvailable;
         FullBufLen = AUDIO_BUF_SIZE * sizeof(int16);
         if(!bEvenRxBuf) {
         // Filling odd buffer, offset to it
            pWriteData += FullBufLen;
         }
      }
      Read = read(p->Socket,pWriteData,FullBufLen-AudioInAvailable);

      if(Read <= 0) {
         if(errno != EAGAIN) {
            LOG_ERROR(("GetAudioIn(): read() failed: %s",Err2String(errno)));
         }
         p->bReadReady = FALSE;
         break;
      }
      else {
         AudioInAvailable += Read;
         if(AudioInAvailable == FullBufLen) {
            AudioInBuf = AudioInBuf16;
            if(!bEvenRxBuf) {
            // just filled the odd buffer point AudioInBuf to it
               AudioInBuf += AUDIO_BUF_SIZE;
            }
            bEvenRxBuf = !bEvenRxBuf;
            if(AudioIn8Bit) {
            // Convert 8 bit audio to 16 bit audio in AudioInBuf
               for(i = 0; i < AUDIO_BUF_SIZE; i++) {
                  AudioInBuf[i] = (int16) ((AudioInBuf8[i] - 128) << 8);
               }
            }
            
            if(bConvertSampleRate) {
            // Convert 48Khz input data to 8Khz
#if 0
               if(fp1 != NULL) {
                  fwrite(AudioInBuf48Khz,AUDIO_BUF_SIZE * 12,1,fp1);
               }
#endif
               DownConvert();
            }
            else if(AudioChannel != 0) {
               int16 *pRight;
               int16 *pLeft;
               int16 *pIn = StereoInBuf;

               if(pAudioSlave != NULL) {
                  pAudioSlave->AudioInBuf = pAudioSlave->AudioInBuf16;
                  if(!pAudioSlave->bEvenRxBuf) {
                  // just filled the odd buffer point AudioInBuf to it
                     pAudioSlave->AudioInBuf += AUDIO_BUF_SIZE;
                  }
                  pAudioSlave->bEvenRxBuf = !pAudioSlave->bEvenRxBuf;

               // Unpack the stereo data into two mono buffers
                  if(AudioChannel == 1) {
                     pRight = AudioInBuf;
                     pLeft = pAudioSlave->AudioInBuf;
                  }
                  else {
                     pRight = pAudioSlave->AudioInBuf;
                     pLeft = AudioInBuf;
                  }

                  for(i = 0; i < AUDIO_BUF_SIZE; i++) {
                     *pLeft++ = *pIn++;
                     *pRight++ = *pIn++;
                  }
                  pAudioSlave->ProcessAudioIn();
               }
               else {
               // Unpack the stereo data into two mono buffers
                  pRight = AudioInBuf;
                  if(AudioChannel == 1) {
                     pIn++;
                  }

                  for(i = 0; i < AUDIO_BUF_SIZE; i++) {
                     *pRight++ = *pIn++;
                     pIn++;
                  }
               }
            }
#if 0
            if(fp != NULL) {
               fwrite(AudioInBuf,sizeof(AudioInBuf),1,fp);
            }
#endif
            ProcessAudioIn();
            AudioInAvailable = 0;
            if(this == pMasterPort) {
            // This port is the master port which acts as a timing reference
            // for all VoIP connections.  Poll all VoIP connections for Tx
            // audio
               struct avl_traverser avl_trans;
               VoipNode *pVoip;
               Node *pNode = (Node *) avl_t_first(&avl_trans,NodeTree);
               while(pNode != NULL) {
                  pVoip = dynamic_cast<VoipNode *> (pNode);
                  if(pVoip != NULL) {
                     pVoip->PollTx();
                  }
                  pNode = (Node *) avl_t_next(&avl_trans);
               }
            }
         }
      }
   }
#endif
}

void GetAudioIn(ClientInfo *p)
{
   CPort *pClass = (CPort *) p->p;

   pClass->GetAudioIn(p);
}

void CPort::PollCOS()
{
   int CosBit = 0;
   int bCos = FALSE;
   int bRx;
   int bCTCSS = FALSE;
   PP_DATA_TYPE IRLP_Bits = 0;
   int ModemCtrl;
#ifdef USB_SUPPORT
   int UsbGPIOData = 0;
#endif

   switch(RxCosMethod) {
      case 2:  // IRLP interface
         if(ClaimIRLP()) {
            LOG_ERROR(("%s: ClaimIRLP failed\n",__FUNCTION__));
         }
         else if(ioctl(IRLPfd,PP_RD_STATUS,&IRLP_Bits) != 0) {
            LOG_ERROR(("%s: ioctl(PP_RD_STATUS) failed: %s",__FUNCTION__,
                       Err2String(errno)));
         }
         else if((IRLP_Bits & COS_Bit)) {
            bCos = TRUE;
         }
         ReleaseIRLP();
         break;

      case 3:  // serial port CTS
         CosBit = TIOCM_CTS;
         break;

      case 4:  // serial port DSR
         CosBit = TIOCM_DSR;
         break;

      case 5:  // serial port DCD
         CosBit = TIOCM_CD;
         break;

      case 6:  // /dev/input/event*
         bCos = bDevInCOS;
         break;

      case 7:  // USB device GPIO
#ifdef USB_SUPPORT
         UsbGPIOData = UsbPollGPIO();
         if(UsbGPIOData & COS_Bit) {
            bCos = TRUE;
         }
#endif
         break;
   }

   if(CosBit != 0) {
      if(ioctl(SerialFd,TIOCMGET,&ModemCtrl) == -1) {
         LOG_ERROR(("PollCOS(): ioctl(TIOCMGET) failed: %s",Err2String(errno)));
      }
      else if(!(ModemCtrl & CosBit)) {
         bCos = TRUE;
      }
   }

   if(InvertCOS) {
      bCos = !bCos;
   }

   if(bCosActive != bCos) {
   // COS has changed state
      bCosActive = bCos;
      DPRINTF(("%s: Port %s COS %sactive\n",__FUNCTION__,NodeName,
               bCos ? "" : "in"));
   }

   switch(RxCtcssMethod) {
      case 1:  // Software CTCSS decoder
         if(bRxCtcssSearch || bRxCtcssAny) {
            if(CurrentRxCTCSS != -1) {
               bCTCSS = TRUE;
            }
         }
         else if(CurrentRxCTCSS == RxCTCSSTone || RxCTCSSTone == -1) {
         // Correct tone or no tone is programmed.
            bCTCSS = TRUE;
         }
         break;

      case 2:  // IRLP interface
         bCTCSS = (IRLP_Bits & CTCSS_Bit) ? TRUE : FALSE;
         break;

      case 3:  // serial port CTS
         if(!(ModemCtrl & TIOCM_CTS)) {
            bCTCSS = TRUE;
         }
         break;

      case 4:  // serial port DSR
         if(!(ModemCtrl & TIOCM_DSR)) {
            bCTCSS = TRUE;
         }
         break;

      case 5:  // serial port DCD
         if(!(ModemCtrl & TIOCM_CD)) {
            bCTCSS = TRUE;
         }
         break;

      case 6:  // /dev/input/event*
         bCTCSS = bDevInCTCSS;
         break;

      case 7:  // USB device GPIO
#ifdef USB_SUPPORT
         if(UsbGPIOData & CTCSS_Bit) {
            bCTCSS = TRUE;
         }
#endif
         break;
   }

   if(InvertCTCSS) {
      bCTCSS = !bCTCSS;
   }

   switch(RxCTCSSMode) {
      default:
         LOG_ERROR(("%s: Unknown RxCTCSSMode %d on port %s, changing to 0\n",
                    __FUNCTION__,RxCTCSSMode,NodeName));
         RxCTCSSMode = 0;
      // Intentional fall through to case 0
      case 0:  // Carrier squelch
         bRx = bCos;
         break;

      case 1:  // Carrier and CTCSS
         bRx = bCos && bCTCSS;
         break;

      case 2:  // Carrier or CTCSS
         if(bRxActive) {
         // Signal active, only need CTCSS to keep squelch open
            bRx = bCTCSS;
         }
         else {
         // Signal not active, need CTCSS and COS to open squelch
            bRx = bCos && bCTCSS;
         }
         break;

      case 3:  // CTCSS only
         bRx = bCTCSS;
         break;

      case 4:  // Carrier and NO CTCSS
         bRx = bCos && !bCTCSS;
         break;
   }

   if(bRx && bCtcssActive != bCTCSS) {
   // CTCSS  has changed state
      bCtcssActive = bCTCSS;
      DPRINTF(("%s: Port %s CTCSS %sactive\n",__FUNCTION__,NodeName,
               bCTCSS ? "" : "in"));
   }

   if(RxTimeout != 0) {
      if(bRx) {
         if(RxActiveTime == 0) {
            RxActiveTime = TimeNow.tv_sec;
         }
         if(!bRxTimeout) {
         // Check for COS timeout
            if(TimeNow.tv_sec - RxActiveTime > RxTimeout) {
               LOG_NORM(("%s: Port %s Rx timeout\n",__FUNCTION__,NodeName));
               bRxTimeout = TRUE;
            }
         }
      }
      else {
         RxActiveTime = 0;
         if(bRxTimeout) {
         // COS Timeout cleared
            LOG_NORM(("%s: Port %s Rx timeout cleared\n",__FUNCTION__,NodeName));
            bRxTimeout = FALSE;
         }
      }

      if(bRxTimeout) {
         bRx = FALSE;
      }
   }

   if(bRxActive != bRx) {
   // Rx active has changed state
      bRxActive = bRx;
      DPRINTF(("%s: Port %s Rx %sactive\n",__FUNCTION__,NodeName,
               bRx ? "" : "in"));

      if(bRx) {
         CtoneTimer.tv_sec = 0;
      }
      else {
         SendCtone(FALSE);
      }

      StartTalking(bRx);
   }
}

void CPort::PollDTMF()
{
   int wrote;
   int BytesRead;
   static char LastDTMF;
   char CurrentDTMF;
   PP_DATA_TYPE IRLP_Bits = 0;

   switch(DtmfMethod) {
      case 1:  // WB2REM / VA3TO interface
         while((BytesRead = read(SerialFd,&CurrentDTMF,1)) == 1) {
            if((CurrentDTMF >= '0' && CurrentDTMF <= '9') ||
               (CurrentDTMF == '*' || CurrentDTMF == '#') || 
               (CurrentDTMF >= 'A' && CurrentDTMF <= 'D') ||
               (CurrentDTMF == ' '))
            {
               if(LastDTMF != CurrentDTMF) {
                  LastDTMF = CurrentDTMF;
                  SaveDTMF(CurrentDTMF);
               }
            }
         }

         if(BytesRead < 0 && errno != EAGAIN) {
            LOG_ERROR(("PollDTMF(): read failed: %s",Err2String(errno)));
         }

         if(bTxIsKeyed != bTxKeyed) {
            bTxIsKeyed = bTxKeyed;
            wrote = write(SerialFd,bTxKeyed ? "T" : "R",1);
         }
         else {
         // don't need to key or unkey, just poll for DTMF
            wrote = write(SerialFd,"C",1);
         }

         if(wrote != 1) {
            LOG_ERROR(("PollDTMF(): write failed: %s",Err2String(errno)));
         }
         break;

      case 2:  // IRLP interface
         if(ClaimIRLP()) {
            LOG_ERROR(("%s: ClaimIRLP failed\n",__FUNCTION__));
         }
         else if(ioctl(IRLPfd,PP_RD_STATUS,&IRLP_Bits) != 0) {
            LOG_ERROR(("%s: ioctl(PP_RD_STATUS) failed: %s",__FUNCTION__,
                       Err2String(errno)));
         }
         else {
            CurrentDTMF = (IRLP_Bits >> IRLP_DTMF_SHIFT) & 0xf;
            if(CurrentDTMF == 0) {
            // IRLP hardware doesn't include the data valid flag from
            // the DTMF decoder, only the 4 data bits.  Since the decoder 
            // used outputs a "D" (0x0) when no digit is decoded it must
            // be tossed since we can't tell if it's valid or not.
               CurrentDTMF = ' ';
            }
            else if(CurrentDTMF < 10) {
               CurrentDTMF += '0';
            }
            else if(CurrentDTMF == 10) {
               CurrentDTMF = '0';
            }
            else if(CurrentDTMF == 11) {
               CurrentDTMF = '*';
            }
            else if(CurrentDTMF == 12) {
               CurrentDTMF = '#';
            }
            else {
               CurrentDTMF = 'A' + (CurrentDTMF - 13);
            }

            if(LastDTMF != CurrentDTMF) {
               LastDTMF = CurrentDTMF;
               SaveDTMF(CurrentDTMF);
            }
         }
         ReleaseIRLP();
         break;
   }
}

int CPort::Audio_Handler(ClientInfo *p)
{
   audio_buf_info BufInfo;
   int bWroteOutput = FALSE;

   bAudioOutCalled = FALSE;

   if(p->bReadReady) {
      GetAudioIn(p);
   }

   if(LastID != 0 && (TimeNow.tv_sec - LastID) > IdInterval) {
   // time to ID
      ID();
   }

   if(bTxKeyed && !bTxTimeOut && TxTimeout != 0 &&  
      (TimeNow.tv_sec - TxKeyTime) > TxTimeout) 
   do {  // Timeout !
      LOG_ERROR(("%s: Port %s Tx timeout\n",__FUNCTION__,NodeName));
      if(TxTimeoutTone != 0) {
         char *ToneSpec = FindToneSpec(TxTimeoutTone);

         if(ToneSpec == NULL) {
            LOG_ERROR(("%s: Timeout tone ID #%d is not defined, disabling.\n",
                       __FUNCTION__,TxTimeoutTone));
            TxTimeoutTone = 0;
            break;
         }

         if((TxTimeoutToneGen = new CToneGen()) == NULL) {
            LOG_ERROR(("%s#%d: new failed\n",__FUNCTION__,__LINE__));
            break;
         }

         if(TxTimeoutToneGen->ToneGenSetup(ToneSpec,NULL) != 0) {
            LOG_ERROR(("%s#%d: ToneGenSetup failed\n",__FUNCTION__,__LINE__));
            delete TxTimeoutToneGen;
            TxTimeoutToneGen = NULL;
         }
      }
      bTxTimeOut = TRUE;
   } while(FALSE);

   if(p->bWriteReady) {
   // Audio output needs another buffer
      WriteWakeups++;
      D2PRINTF(("R"));
      bWroteOutput = WriteAudioOut(p);
   }


   PollCOS();

   if(DtmfMethod == 2 ||
      (DtmfMethod == 1 && TimeLapse(&LastDTMFPoll) >= DTMFPollTime))
   {
      PollDTMF();
      LastDTMFPoll = TimeNow;
   }


   if(CtoneTimer.tv_sec != 0) {
      if(TimeLapse(&CtoneTimer) >= CtoneDelay) {
         SendCtone(TRUE);
         CtoneTimer.tv_sec = 0;
      }
   }

   if(!bAudioOutCalled && pPlayHead != NULL &&
      (!(pPlayHead->Flags & PLAY_FLAG_TTS_OUT) || !bTTS_Running))
   {  // Output audio wasn't running and there's an announcement to play
      // and it's either not an TTS rendered announcement or the TTS
      // system has finished rendering it...  start the output audio stream
      bWroteOutput = WriteAudioOut(pAudioC);
   }

#if 0 // !!! fix this
   if(!bTxKeyed && bTxTimeOut) {
      if((RxMap & ActiveRxMap) == 0 && !ActivePorts()) {
         LOG_NORM(("%s: Cleared tx timeout\n",__FUNCTION__));
         bTxTimeOut = 0;
      }
   }
#endif

   if(!bAudioOutCalled 
#if 0 // !!! maybe??
      && SourcesReady(NULL)
#endif
      
      ) {
      bWroteOutput = WriteAudioOut(pAudioC);
   }

   if(bTxKeyed && !bWroteOutput) {
   // Check to see if the output buffer has drained to zero
      if(ioctl(pAudioC->Socket,SNDCTL_DSP_GETOSPACE,&BufInfo) == -1) {
         LOG_ERROR(("Audio_Handler(): SNDCTL_DSP_GETOSPACE failed: %s",
                    Err2String(errno)));
      }
      else if(BufInfo.fragments == BufInfo.fragstotal) {
      // The output buffer is empty, unkey the transmitter
         if(HangTime == 0) {
         // Transmitter has no hang time, unkey now
            KeyTx(FALSE);
         }
         else if(HangTimer.tv_sec == 0) {
         // Start HangTimer
            HangTimer = TimeNow;
         }
         else if(TimeLapse(&HangTimer) >= HangTime) {
            HangTimer.tv_sec = 0;
            KeyTx(FALSE);
         }
      }
   }

   DFLUSH();

   return FALSE;
}

int Audio_Handler(ClientInfo *p)
{
   CPort *pClass = (CPort *) p->p;

   return pClass->Audio_Handler(p);
}

void CPort::AudioCleanup(ClientInfo *p)
{
   KeyTx(FALSE);
   
   if(TxKeyMethod == 1 && SerialFd != -1) {
   // WB2REM / VA3TO interface
      usleep(1000*DTMFPollTime);
      write(SerialFd,"R",1);
   }
}

void AudioCleanup(ClientInfo *p)
{
   CPort *pClass = (CPort *) p->p;
   pClass->AudioCleanup(p);
}

int CPort::SerialInit()
{
   int Ret = 0;
   struct termios options;
   int i;
   
   if(SerialDevName == NULL) {
      LOG_ERROR(("SerialInit(): Error SerialDevName is not defined.\n"));
      Ret = ERR_SERIAL_DEV_OPEN;
   }
   else if((SerialFd = open(SerialDevName,SERIAL_OPEN_FLAGS)) < 0) {
      LOG_ERROR(("SerialInit(): Couldn't open \"%s\": %s",SerialDevName,
                 Err2String(errno)));
      Ret = ERR_SERIAL_DEV_OPEN;
   }
   else if(tcgetattr(SerialFd,&options) < 0) {
      LOG_ERROR(("SerialInit(): Unable to read port configuration: %s",
                 Err2String(errno)));
      Ret = ERR_SERIAL_GET_ATTR;
   }
   else {
      options.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);    /* raw input */
      options.c_iflag &= ~(IXON | IXOFF | IXANY);  /* disable software flow control */
      options.c_iflag |=  IGNBRK;            /* ignore break */
      options.c_iflag &= ~ISTRIP;            /* do not strip high bit */
      options.c_iflag &= ~(INLCR | ICRNL);   /* do not modify CR/NL   */
      options.c_cflag |=  (CLOCAL | CREAD);  /* enable the receiver and set local mode */
      options.c_cflag &= ~CRTSCTS;           /* disable hardware flow control */
      options.c_cflag &= ~CSIZE;             /* Mask the character size bits */
      options.c_cflag |= CS8;                /* 8 data bits */
      options.c_cflag &= ~PARENB;            /* no parity */
      options.c_cflag &= ~CSTOPB;            /* 1 stop bit */

      for(i = 0; BaudRateTbl[i].Baudrate != 0; i++) {
         if(BaudRateTbl[i].Baudrate == SerialBaudRate) {
            if(cfsetspeed(&options,BaudRateTbl[i].Baudrate) == -1) {
               LOG_ERROR(("SerialInit(): cfsetspeed() failed: %s",
                          Err2String(errno)));
               Ret = ERR_SERIAL_SET_BAUD;
            }
            break;
         }
      }

      if(BaudRateTbl[i].Baudrate == 0) {
         LOG_ERROR(("SerialInit(): Baudrate %d not supported.\n"));
         Ret = ERR_SERIAL_SET_BAUD;
      }
   }

   if(Ret == 0 && tcsetattr(SerialFd,TCSANOW,&options) < 0) {
      LOG_ERROR(("SerialInit(): tcsetattr() failed: %s",Err2String(errno)));
      Ret = ERR_SERIAL_SET_ATTR;
   }

   if(Ret != 0) {
      close(SerialFd);
      SerialFd = -1;
   }

   return Ret;
}

void CPort::SendCWMsg(char *Msg)
{
   CWChar = Msg;
   if(!pAudioC->bWriteWait) {
      WriteAudioOut(pAudioC);
   }
}

int CPort::AudioInit()
{
   int Ret = 0;
   int SampleSize = 16;
   int SampleRate;
   int Stereo = 0;
   int Fmt = AFMT_S16_LE;
   int bits;
   audio_buf_info BufInfo;
#ifdef SNDCTL_DSP_SETBLKSIZE
   int BlkSz;
#else
   long frag_size;
#endif

   do {
      if(AudioIn8Bit) {
         if((AudioInBuf8 = (char *) malloc(AUDIO_BUF_SIZE)) == NULL) {
            LOG_ERROR(("%s#%d: malloc failed\n",__FUNCTION__,__LINE__));
            Ret = ERR_MALLOC;
            break;
         }
      }

      if((AudioInBuf16 = (int16 *) malloc(AUDIO_BUF_SIZE*4)) == NULL) {
         LOG_ERROR(("%s#%d: malloc failed\n",__FUNCTION__,__LINE__));
         Ret = ERR_MALLOC;
         break;
      }

      AudioInBuf = AudioInBuf16;

      if(AudioChannel == 0) {
      // Open audio port in Mono mode
         Stereo = 0;
#ifdef SNDCTL_DSP_SETBLKSIZE
         BlkSz = 1280;
#else
         frag_size = 0x50008;  // 5 x 256 byte fragments
#endif
      }
      else {
      // Check to see if this audio device has been been opened by a prior
      // port
         Ret = FindAudioMaster();
         if(Ret == -1) {
         // We're a slave of another port, clear the error and return
            Ret = 0;
            break;
         }
         else if(Ret != 0) {
         // Error detected
            break;
         }

      // Open audio port in Stereo mode
         Stereo = 1;
#ifdef SNDCTL_DSP_SETBLKSIZE
         BlkSz = 1280*2;
#else
         frag_size = 0xa0008;  // 10 x 256 byte fragments
#endif
         if((StereoInBuf = (int16 *) malloc(AUDIO_BUF_SIZE*8)) == NULL) {
            LOG_ERROR(("%s#%d: malloc failed\n",__FUNCTION__,__LINE__));
            Ret = ERR_MALLOC;
            break;
         }

         if((StereoOutBuf = (int16 *) malloc(AUDIO_BUF_SIZE*8)) == NULL) {
            LOG_ERROR(("%s#%d: malloc failed\n",__FUNCTION__,__LINE__));
            Ret = ERR_MALLOC;
            break;
         }
      }

#ifdef USB_SUPPORT
      if(UsbDevType == USB_TYPE_CM108) {
         if(Stereo) {
            LOG_ERROR(("%s: Invalid configuration, the CM108 chip does not "
                       "support stereo input\n",__FUNCTION__));
            Ret = ERR_CONFIG_FILE;
            break;
         }

         LOG_ERROR(("%s: Internal sampling rate conversion for CM108 %sabled\n",
                    __FUNCTION__,
                    DisableRateConversion ? "dis" : "en"));

         if(!DisableRateConversion) {
            bConvertSampleRate = TRUE;
            PCMRate = 48000;
            if((AudioBuf48Khz = (int16 *) malloc(AUDIO_BUF_SIZE*12)) == NULL) {
               LOG_ERROR(("%s#%d: malloc failed\n",__FUNCTION__,__LINE__));
               Ret = ERR_MALLOC;
               break;
            }

            if((AudioInBuf48Khz = (int16 *) malloc(AUDIO_BUF_SIZE*12)) == NULL) {
               LOG_ERROR(("%s#%d: malloc failed\n",__FUNCTION__,__LINE__));
               Ret = ERR_MALLOC;
               break;
            }
         }
      }
#endif

      pAudioC = CreateNewClient();

      if(pAudioC == NULL) {
         Ret = ERR_MALLOC;
         break;
      }

      pAudioC->Cleanup = ::AudioCleanup;
      pAudioC->State = ::Audio_Handler;
      pAudioC->p = this;

      LOG_ERROR(("%s#%d: opening \"%s\"\n",__FUNCTION__,__LINE__,AudioDevice));

      if((pAudioC->Socket = open(AudioDevice,AUDIO_OPEN_FLAGS)) < 0) {
         LOG_ERROR(("%s#%d: open(\"%s\") failed: %s",__FUNCTION__,__LINE__,AudioDevice,
                   Err2String(errno)));
         Ret = ERR_AUDIO_DEV_OPEN;
         break;
      }

      if(ioctl(pAudioC->Socket,SNDCTL_DSP_SETDUPLEX,NULL) == -1) {
      // Failure is not necessarily fatal.
      // FreeBSD doesn't require this call for fullduplex
         LOG_WARN(("EndPointInit(): SNDCTL_DSP_SETDUPLEX failed: %s",
                   Err2String(errno)));
      }

      if(ioctl(pAudioC->Socket,SNDCTL_DSP_GETFMTS,&bits) == -1) {
         LOG_WARN(("EndPointInit(): SNDCTL_DSP_GETFMTS failed: %s",
                   Err2String(errno)));
      }
      else if(!(bits & AFMT_S16_LE)) {
         LOG_ERROR(("EndPointInit(): Error AFMT_S16_LE not supported\n"));
         Ret = ERR_AUDIO_FMT;
         break;
      }
      
      if(ioctl(pAudioC->Socket,SNDCTL_DSP_SAMPLESIZE,&SampleSize) == -1) {
         LOG_ERROR(("EndPointInit(): SNDCTL_DSP_SAMPLESIZE failed: %s",
                    Err2String(errno)));
         Ret = ERR_AUDIO_FMT;
         break;
      }
      
      if(ioctl(pAudioC->Socket,SNDCTL_DSP_STEREO,&Stereo) == -1) {
         LOG_ERROR(("EndPointInit(): SNDCTL_DSP_STEREO failed: %s",
                    Err2String(errno)));
         Ret = ERR_AUDIO_FMT;
         break;
      }
      
      SampleRate = PCMRate;
      if(ioctl(pAudioC->Socket,SNDCTL_DSP_SPEED,&SampleRate) == -1) {
         LOG_ERROR(("EndPointInit(): SNDCTL_DSP_SPEED failed: %s",
                    Err2String(errno)));
         Ret = ERR_AUDIO_FMT;
         break;
      }
      else if(SampleRate != PCMRate) {
         LOG_WARN(("%s: Asked for a sampling rate of %d, got %d\n",
                   __FUNCTION__,PCMRate,SampleRate));
      }
      
      if(ioctl(pAudioC->Socket,SNDCTL_DSP_SETFMT,&Fmt) == -1) {
         LOG_ERROR(("EndPointInit(): SNDCTL_DSP_SETFMT failed: %s",
                    Err2String(errno)));
         Ret = ERR_AUDIO_FMT;
         break;
      }

#ifdef SNDCTL_DSP_SETBLKSIZE
      if(ioctl(pAudioC->Socket,SNDCTL_DSP_SETBLKSIZE,&BlkSz) == -1) {
         LOG_ERROR(("EndPointInit(): SNDCTL_DSP_SETBLKSIZE failed: %s",
                    Err2String(errno)));
         Ret = ERR_AUDIO_FMT;
         break;
      }
#else
      if(ioctl(pAudioC->Socket,SNDCTL_DSP_SETFRAGMENT,&frag_size) == -1) {
         LOG_ERROR(("EndPointInit(): SNDCTL_DSP_SETFRAGMENT failed: %s",
                    Err2String(errno)));
         Ret = ERR_AUDIO_SET_FRAG;
         break;
      }
#endif
      
      if(ioctl(pAudioC->Socket,SNDCTL_DSP_GETOSPACE,&BufInfo) == -1) {
         LOG_ERROR(("EndPointInit(): SNDCTL_DSP_GETOSPACE failed: %s",
                    Err2String(errno)));
      }
      else {
         LOG_ERROR(("Fragsize: %d, fragstotal: %d, bytes: %d\n",
                    BufInfo.fragsize,BufInfo.fragstotal,BufInfo.bytes));
      }
   } while(FALSE);

   return Ret;
}

int CPort::EndPointInit()
{
   int Ret = 0;

   LOG_NORM(("%s: called, TxKeyMethod: %d, RxCosMethod: %d\n",__FUNCTION__,
             TxKeyMethod,RxCosMethod));

   bEchoIrlpMode = TRUE;

   for( ; ; ) {
      if(AudioIn8Bit) {
         if((AudioInBuf8 = (char *) malloc(AUDIO_BUF_SIZE)) == NULL) {
            LOG_ERROR(("%s#%d: malloc failed\n",__FUNCTION__,__LINE__));
            Ret = ERR_MALLOC;
            break;
         }
      }

      if((AudioInBuf16 = (int16 *) malloc(AUDIO_BUF_SIZE*4)) == NULL) {
         LOG_ERROR(("%s#%d: malloc failed\n",__FUNCTION__,__LINE__));
         Ret = ERR_MALLOC;
         break;
      }

      AudioInBuf = AudioInBuf16;

      if(TxKeyMethod == 5) {
#ifdef USB_SUPPORT
         if((Ret = UsbInit()) != 0) {
            break;
         }
#else
         LOG_ERROR(("%s: Error - thelinkbox was not configured with USB support.\n",
                    __FUNCTION__));
         Ret = ERR_USB_DEV_OPEN;
         break;
#endif
      }

      if(RxCosMethod == 6) {
#ifdef __linux__
         if((Ret = InitDeviceInput()) != 0) {
            LOG_ERROR(("%s: Couldn't find /dev/inputX entry for RxCosMethod 6\n",
                       __FUNCTION__));
            break;
         }
#else
         LOG_ERROR(("Error: RxCosMethod is only supported on Linux\n"));
         Ret = ERR_INPUT_DEV_OPEN;
         break;
#endif
      }

      if((Ret = AudioInit()) != 0) {
         break;
      }
      
      if((TxKeyMethod == 1 || TxKeyMethod == 3 || TxKeyMethod == 4 ||
          RxCosMethod == 3 || RxCosMethod == 4 || RxCosMethod == 5))
      {  // serial port needed by TxKeyMethod or RxCosMethod

         if(TxKeyMethod == 3) {
            PTT_Bit = TIOCM_RTS;
         }
         else if(TxKeyMethod == 4) {
            PTT_Bit = TIOCM_DTR;
         }

         if(!FindSerialMaster()) {
         // This serial port has not been opened yet
            if((Ret = SerialInit()) != 0) {
               break;
            }
         }
      }

      if(TxKeyMethod == 2 || RxCosMethod == 2) {
      // IRLP port needed by TxKeyMethod or RxCosMethod
         if(PTT_Bit == -1) {
            PTT_Bit = IRLP_PTT;
         }

         if(COS_Bit == -1) {
            COS_Bit = IRLP_COS;

         }
         if((IRLPfd = open(IRLP_DEVICE,O_RDWR)) < 0 ) { 
            LOG_ERROR(("EndPointInit(): open failed for \"%s\": %s",
                       IRLP_DEVICE,Err2String(errno)));
            Ret = ERR_IRLP_DEV_OPEN;
            break;
         }

         if(ShareIRLP_Port) {
            if((LockIRLPfd = open(IRLP_LOCKFILE,O_WRONLY|O_CREAT, 0664)) < 0 ) { 
               LOG_ERROR(("EndPointInit(): open failed for \"%s\": %s",
                          IRLP_LOCKFILE,Err2String(errno)));
               Ret = ERR_IRLP_LOCKFILE_OPEN;
               break;
            }
         }
         else {
#ifdef __linux__
            if(ioctl(IRLPfd,PPCLAIM) != 0) {
                LOG_ERROR(("EndPointInit(): ioctl(PPCLAIM) failed: %s",
                           Err2String(errno)));
                Ret = ERR_IRLP_DEV_CLAIM;
                break;
            } 
#endif
         }

         if(InitPPData != -1) {
            char  IRLP_Bits = (char) InitPPData;

            if(ClaimIRLP()) {
               LOG_ERROR(("%s: ClaimIRLP failed\n",__FUNCTION__));
            }
            else {
               if(ioctl(IRLPfd,PP_WR_DATA,&IRLP_Bits)) {
                  LOG_ERROR(("%s: ioctl(PPWDATA) failed: %s",__FUNCTION__,
                             Err2String(errno)));
               }
               ReleaseIRLP();
            } 
         }

         LOG_NORM(("%s: IRLPfd: %d\n",__FUNCTION__,IRLPfd));
      }

      if(DtmfMethod == 3) {
         dtmf_rx_init(&DtmfRxState,NULL,this);
         dtmf_rx_set_realtime_callback(&DtmfRxState,DtmfCallback,this);
         dtmf_rx_parms(&DtmfRxState,0,Twist,ReverseTwist);
      }

      avl_insert(ClientTree,pAudioC);
      pAudioC->bInClientTree = TRUE;
      pAudioC->bReadWait = TRUE;
      GetAudioIn(pAudioC);
      break;
   }

   if(Ret != 0) {
   // Init error, clean up
      if(pAudioC != NULL) {
         if(pAudioC->Socket >= 0) {
            close(pAudioC->Socket);
         }
         DeleteClient(pAudioC);
         pAudioC = NULL;
      }

      if(IRLPfd != -1) {
         close(IRLPfd);
         IRLPfd = -1;
      }

      if(LockIRLPfd != -1) {
         close(LockIRLPfd);
         LockIRLPfd = -1;
      }
   }
   else {
      KeyTx(FALSE);
   }


   LOG_NORM(("%s: returning %d\n",__FUNCTION__,Ret));

   return Ret;
}

int EndPointInit() 
{
   int Ret = -1;
   struct avl_traverser avl_trans;
   CPort *pPort;
   Node *pNode;

#if 0
   fp = fopen("test.raw","w");
   fp1 = fopen("out.raw","w");
#endif

   pNode = (Node *) avl_t_first(&avl_trans,NodeTree);
   while(pNode != NULL) {
      pPort = dynamic_cast<CPort *> (pNode);
      if(pPort != NULL) {
         LOG_ERROR(("%s: Initializing %s port\n",__FUNCTION__,pPort->NodeName));
         if((Ret = pPort->EndPointInit()) != 0) {
            break;
         }
         if(pMasterPort == NULL) {
            pMasterPort = pPort;
            CurrentNode = pNode;
         }
      }
      pNode = (Node *) avl_t_next(&avl_trans);
   }
   return Ret;
}

void CPort::CmdForceID()
{
   SendCWMsg(CWId);
}

void SetTimeout(struct timeval *pTimer,int Timeout)
{
   long microseconds;

   microseconds = TimeNow.tv_usec + ((Timeout % 1000) * 1000);

   pTimer->tv_usec = microseconds % 1000000;
   pTimer->tv_sec  = TimeNow.tv_sec + (Timeout / 1000) + microseconds / 1000000;
}

// Try to find a previously initialized port that's using the same
// serial port as this port
int CPort::FindSerialMaster()
{
   int Ret = FALSE;
   return Ret;
}

void CmdForceID(ClientInfo *p,ConfClient *pCC1,char *Arg)
{
   CPort *pPort = dynamic_cast<CPort *> (CurrentNode);

   if(pPort != NULL) {
      pPort->CmdForceID();
   }
}

void CPort::DumpStatus(ClientInfo *p)
{
   int bFirst = TRUE;

   BufPrintf(p,"%c %s",CurrentNode == this ? '>' : ' ',NodeName);
#if 0 // !!!
   if(ActiveRxMap & MyBit) {
      BufPrintf(p,"%sRx",bFirst ? " " : ", ");
      bFirst = FALSE;
   }
#endif
   if(bTxKeyed) {
      BufPrintf(p,"%sTx",bFirst ? " " : ", ");
      bFirst = FALSE;
   }
   if(bTxTimeOut) {
      BufPrintf(p,"%sTxTo",bFirst ? " " : ", ");
      bFirst = FALSE;
   }
   if(MuteDTMF && DTMF_Digits != 0) {
      BufPrintf(p,"%sDtmfMute",bFirst ? " " : ", ");
      bFirst = FALSE;
   }
   if(bRxTimeout) {
      BufPrintf(p,"%sRxTo",bFirst ? " " : ", ");
      bFirst = FALSE;
   }

   BufPrintf(p,"\r");
}

void CPort::AddTx(LinkedNode *pNewLink)
{
   Node::AddTx(pNewLink);
   pNewLink->BufNum = BufNum;
}

// Return 0 to continue audio port initialization,
// -1 if we're a slave of another port or error code.
int CPort::FindAudioMaster()
{
   struct avl_traverser avl_trans;
   CPort *pPort;
   Node *pNode;
   int Ret = 0;
   int PortsFound = 0;

   pNode = (Node *) avl_t_first(&avl_trans,NodeTree);
   while(pNode != NULL) {
      pPort = dynamic_cast<CPort *> (pNode);
      if(pPort != NULL && pPort != this && 
         strcmp(AudioDevice,pPort->AudioDevice) == 0) 
      {
      // Same audio device
         PortsFound++;
         if(pPort->AudioChannel == 0) {
         // Other port is configured as mono
            LOG_ERROR(("%s: Invalid audio port sharing, port %s is configured "
                       "for mono\n",__FUNCTION__,pPort->NodeName));
            Ret = ERR_AUDIO_DEV_OPEN;
            break;
         }
         else if(pPort->AudioChannel == AudioChannel) {
         // Other port is configured as for the same channel
            LOG_ERROR(("%s: Invalid audio port sharing, ports %s and %s are "
                       "configured for the same channel\n",__FUNCTION__,
                       NodeName,pPort->NodeName));
            Ret = ERR_AUDIO_DEV_OPEN;
            break;
         }
         else {
            if(pPort->pAudioC != NULL) {
            // We're his slave
               pPort->pAudioSlave = this;
               Ret = -1;
            }
         }
      }
      pNode = (Node *) avl_t_next(&avl_trans);
   }

   if(PortsFound > 1) {
      LOG_ERROR(("%s: Invalid audio port sharing, more than two ports share "
                 "%s\n",__FUNCTION__,pPort->AudioDevice));
      Ret = ERR_AUDIO_DEV_OPEN;
   }
   return Ret;
}

int CPort::SetSoftRxTone(int ToneNum)
{
   int Ret = 0;

   do {
      bRxCtcssSearch = bRxCtcssAny = FALSE;
      RxCTCSSTone = ToneNum;
      if(ToneNum == -1) {
      // Disable CTCSS
         if(CtcssDecoderState != NULL)  {
            CtcssDestroy(CtcssDecoderState);
            CtcssDecoderState = NULL;
         }
         if(RxToneDesc != NULL) {
            free(RxToneDesc);
            RxToneDesc = strdup("disabled");
         }
      }
      else {
         if(CtcssDecoderState == NULL) {
            if((CtcssDecoderState = CtcssInit()) == NULL) {
               LOG_ERROR(("%s: CtcssInit() failed\n",__FUNCTION__));
               Ret = ERR_CTCSS_RX_INIT;
               break;
            }
         }
         EnableCtcssTone(-1,CtcssDecoderState); // Disable all tones
      // Enable the selected tone plus the guard tones
         if(ToneNum != 0) {
            EnableCtcssTone(ToneNum - 1,CtcssDecoderState);
         }
         if(ToneNum + 1 < CTCSS_NUM_CODES) {
            EnableCtcssTone(ToneNum + 1,CtcssDecoderState);
         }
         EnableCtcssTone(ToneNum,CtcssDecoderState);
         if(RxToneDesc != NULL) {
            free(RxToneDesc);
            RxToneDesc = strdup(freq_ctcss[ToneNum]);
         }
      }
   } while(FALSE);
   return Ret;
}

// 
int CPort::SetRxTone(char *Arg)
{
   int CtcssFreq = -99;
   float CtcssFloat;
   int i;
   int Ret = ERR_CONFIG_FILE;    // assume the worse
   int bToneFreqSet = FALSE;

   do {
      if(RxCtcssMethod == 0) {
         LOG_ERROR(("%s: No RxCtcssMethod specified for port %s, "
                    "RxTone ignored\n",__FUNCTION__,NodeName));
         break;
      }

      if(RxCtcssMethod == 1) {
      // the software decoder can do some special tricks!
         int bSearch = FALSE;
         int bAny = FALSE;
         if((bAny = (STRCMP(Arg,"any") == 0)) || 
            (bSearch = (STRCMP(Arg,"search") == 0)))
         {
            if(CtcssDecoderState == NULL) {
               if((CtcssDecoderState = CtcssInit()) == NULL) {
                  LOG_ERROR(("%s: CtcssInit() failed\n",__FUNCTION__));
                  Ret = ERR_CTCSS_RX_INIT;
                  break;
               }
            }
            if(RxToneDesc != NULL) {
               free(RxToneDesc);
               RxToneDesc = strdup(Arg);
            }
            EnableCtcssTone(-2,CtcssDecoderState); // enable all tones
            bRxCtcssSearch = bSearch;
            bRxCtcssAny = bAny;
            Ret = 0;
            break;
         }
      }

      if(strchr(Arg,'.') != NULL) {
      // There's a decimal point, must be a CTCSS frequency
         if(sscanf(Arg,"%f",&CtcssFloat) == 1) {
            bToneFreqSet = TRUE;
            CtcssFreq = (int) ((CtcssFloat * 10.0) + .5);
         }
      }
      else {
      // decimal (freq x 10) CTCSS frequency or (eventually) DCS code
         if(sscanf(Arg,"%d",&CtcssFreq) == 1) {
            bToneFreqSet = TRUE;
         }
      }

      if(!bToneFreqSet) {
         LOG_ERROR(("%s: Unable to parse RxTone \"%s\" for Port %s\n",
                    __FUNCTION__,Arg,NodeName));
         break;
      }

      if(CtcssFreq == 0) {
      // disable CTCSS
         i = -1;
      }
      else {
      // Find Ctcss tone number
         for(i = 0; i < CTCSS_NUM_CODES; i++) {
            if(CtcssFreq == Freq2ToneNum[i]) {
            // Found it
               break;
            }
         }
      }

      if(i == CTCSS_NUM_CODES) {
         LOG_ERROR(("%s: Invalid RxTone \"%s\" for Port %s ignored\n",
                    __FUNCTION__,Arg,NodeName));
         break;
      }

      if(RxCtcssMethod == 1) {
         Ret = SetSoftRxTone(i);
      }
      else {
         LOG_ERROR(("%s: RxCtcssMethod %d for port %s, does not support "
                    "RxTone selection\n",__FUNCTION__,RxCtcssMethod,NodeName));

      }
   } while(FALSE);
   return Ret;
}


