/* Copyright (C) 2002 GFRN systems

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The latest version of this program may be found at
   http://CQiNet.sourceforge.net

   $Log: ports.h,v $
   Revision 1.11  2008/04/25 05:23:23  wb6ymh
   1. Began modifications to allow a single stereo sound device to be use to
      interface to two separate radios.
   2. Modified the code to use 48khz to talk CM108 based USB dongles.  Added
      local sampling rate conversion routines to convert between 8Khz and 48Khz.
   3. Added support for site prefix decoding.
   4. Removed unneeded ClientInfo pointer arguments to ProcessAudioIn, PollCOS,
      and PollDTMF.
   5. Completed support for software based CTCSS decoding including search
      and decode all tones options.
   6. Added SetRxTone function.

   Revision 1.10  2008/03/11 14:12:13  wb6ymh
   Added AddTx override.

   Revision 1.9  2008/03/09 17:41:54  wb6ymh
   Deleted CWToneFreq.

   Revision 1.8  2008/03/09 17:03:53  wb6ymh
   Massive restructuring: moved most CPort members to new Nodes class. All
   that remains are variables specific to physical ports.

   Revision 1.7  2008/01/28 15:39:59  wb6ymh
   1. Added support for external CTCSS decoders controlled by configuration
      variables RxCtcssMethod, RxCtCSSMode, InverCTCSS, and DtfmCTCSSMode.
   2. Corrected bug in CPort::Shutup that prevented it from ever doing anything.

   Revision 1.6  2008/01/13 17:06:17  wb6ymh
   1. Added RxTimeout, TxTimeoutTone, DtmfConverTone, bRxTimeout, bEvenRxBuf,
      TossRxIn, AudioInBuf15, RxActiveTime.
   2. Modified CancelToneGen to return status.
   3. Added bDelayed argument to SendCtone.

   Revision 1.5  2007/12/27 17:58:10  wb6ymh
   1. Added DumpStatus class member.
   2. Added MinDtmfInterdigitTime, Min1stDigitHoldTime and Dtmf1stDigits
      variables.

   Revision 1.4  2007/12/14 23:28:21  wb6ymh
   Moved globals AudioBuf, AudioAvailable, TxSequency, RtpTxSequence,
   DtmfRxState and all GSM variables into CPort class.

   Revision 1.3  2007/12/07 16:57:23  wb6ymh
   1. Moved global ADPCM variables into CPort class.
   2. Added prototypes for CancelToneGen, SourcesReady, and GetClientAudio.
   3. Added TxTimeoutToneGen, CtoneResetsTimeout, and b8BitFile variables.

   Revision 1.2  2007/12/01 06:03:49  wb6ymh
   1. Added include for new ToneGen.h.
   2. Deleted GenTestSweep.
   3. Added class members and variables to implement courtesy tone, hangtime
      and transmitter timeout functions.

   Revision 1.1  2007/11/26 15:19:55  wb6ymh
   Inital import - header for CPort class.

*/

#ifndef _PORTS_H_
#define _PORTS_H_


#define MAX_PORTS 32


class CPort : public Node
{
public:
// public functions   
   CPort();
   ~CPort();
// entry points from C code
   int EndPointInit();
   int WriteAudioOut(ClientInfo *p);
   int Audio_Handler(ClientInfo *p);
   void GetAudioIn(ClientInfo *p);
   void AudioCleanup(ClientInfo *p);
   int DeviceInHandler(ClientInfo *p);
   void CmdForceID();
   void DumpStatus(ClientInfo *p);
   int PausePlayback();
   int ResumePlayback();
   int GetRxAudio(LinkedNode *pLink,int16 **pAudio);
   void AddTx(LinkedNode *pLink);
   int SetRxTone(char *Arg);

// Public variables

// variables accessed by Node class
   ClientInfo *pAudioC;

// Config variables
   int VoxThreshold;
   int VoxHoldTime;
   int VoxTripDelay;
   int TxKeyMethod;
   int RxCosMethod;
   int RxCtcssMethod;
   int SerialBaudRate;
   int DTMFPollTime;
   int InvertPTT;
   int InvertCOS;
   int InvertCTCSS;
   int IdInterval;
   int VoiceIDIdleTime;
   int HalfDuplexRF;
   int HalfDuplex;
   int PCMRate;
   int CtcssDecoderSamples;
   int bMonitor;
   int COS_KeyCode;
   int AudioIn8Bit;
   int bDiscFeed;
   int bUngatedRx;
   int PTT_Bit;
   int COS_Bit;
   int CTCSS_Bit;
   int HangTime;
   int RxCTCSSMode;
   int CTCSS_KeyCode;
   int AudioChannel;
   int RxCTCSSTone;  // Tone index or -1 for carrier
   int CtcssRxThreshold;
   int DisableRateConversion;

   float RxFrequency;
   float TxTone;
   float RxTone;
   float TxOffset;

   char *SerialDevName;
   char *CWId;
   char *VoiceID;
   char *AudioDevice;
   char *AudioDongleSN;
   char *AudioDongleAdr;
   char *RxToneDesc;

private:
   struct timeval LastCOSPoll;
   struct timeval LastSpoke;
   struct timeval LastVoxTrip;
   struct timeval LastDTMFPoll;
   struct timeval HangTimer;
   struct timeval CtcssDecodeTime;
   time_t LastID;
   int bTxKeyed;
   int bTalking;     // receiving signal from RF user
   int bTxIsKeyed;
   int bDevInCOS;
   int bDevInCTCSS;
   int bEvenRxBuf;
   int TossRxIn;
   int bCosActive;

   int AudioInAvailable;
   char *AudioInBuf8;
   int16 *AudioInBuf;
   int16 *AudioInBuf16;
   int16 *StereoInBuf;
   int16 *StereoOutBuf;
   int16 *AudioBuf48Khz;
   int16 *AudioInBuf48Khz;
   CPort *pSerialSlave;
   CPort *pParallelSlave;
   CPort *pAudioSlave;
   int SerialFd;
   void *CtcssDecoderState;
   int CurrentRxCTCSS;
   int DCTemp[4];
   int UCTemp[4];
   int CtcssTemp[4];
   int16 UCLastSample;

   time_t RxActiveTime;

   int AudioUsed;
   long int PausedOffset;
   char *PausedFilename;
#ifdef USB_SUPPORT
   usb_dev_handle *udev;
   int UsbDevType;
#endif
// Kludge... fix me!
   int bAudioOutCalled;
   int bRxCtcssSearch;
   int bRxCtcssAny;
   int bConvertSampleRate:1;

   void KeyTx(int bKey);
   void ID();
   void FillAudioBuf();
   void Vox();
   void ProcessAudioIn();
   void PollCOS();
   void PollDTMF();
   int SerialInit();
   void FillCTCSSBuf();
   int InitDeviceInput();
   int UsbInit(void);
   void StartTalking(int bTalk);
   void SendCWMsg(char *Msg);
   void UsbKeyRadio(int bKey);
   int UsbPollGPIO();
   int ActivePorts();
   int SourcesReady(ConfClient **pTalker);
   int FindSerialMaster(void);
   int VoipAudioAvailable(ConfClient *pCC);
   int GetClientAudio(ConfClient *pCC);
   int AudioInit(void);
   int FindAudioMaster();
   void DownConvert(void);
   void UpConvert(void);
   int SetSoftRxTone(int ToneNum);
};




extern Node *CurrentNode;
extern struct avl_table *NodeTree;   // sorted by node name

// bit map of ports that have active receive signals
extern int ActiveRxMap;
#endif   // _PORTS_H_

