/* Copyright (C) 2008 GFRN systems

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The latest version of this program may be found at
   http://CQiNet.sourceforge.net

   $Log: voip.h,v $
   Revision 1.2  2008/03/11 14:22:26  wb6ymh
   1. Moved RtpData here from conference.h.
   2. Added AllocAudioBuf.

   Revision 1.1  2008/03/09 17:10:14  wb6ymh
   Inital import - files broken out of old audio.cpp and audio.h due to
   major restructuring of the internal infrastructure to treat VoIP connections
   as pseudo ports.

*/
#ifndef _VOIP_H_
#define _VOIP_H_

typedef struct RtpDataTAG {
   struct RtpDataTAG *Link;   // NB: must be first
   rtp_hdr_t   *pData;
   int   Len;           // length of packet
} RtpData;

class VoipNode : public Node
{
public:
// public functions   
   VoipNode(ConfClient *pCC);
   ~VoipNode();
   int WriteAudioOut(ClientInfo *p);
   int GetRxAudio(LinkedNode *pLink,int16 **pAudio);
   void QueueAudio(NetAudio *pAudio);
   void StopTalking();
   void DecompressAudio();
   void QueueSFData(soundbuf *p);
   void QueueRTPData(RtpData *pRtp);
   void EndPointCleanup();
   void PollTx();
   static void VoipInit(ConfClient *pCC);
   int Enabled();

   ConfClient *mCC;

private:
   int   MissedPackets;
   int   OutofSeqPackets;
   int   DuplicatePackets;
   int   LatePackets;

// sequence number of next packet that will be converted to audio when needed
   u_int16  NextSeq;

// Queue of audio packets
   NetAudio *pInHead;
   NetAudio *pInTail;
   int      AudioQCount;

   RtpData  *pRtpQueue;
   int   RtpQCount;
   gsm   gsmh;

// receive state transisions: 
// idle -> buffering -> active -> idle
//           ^            |
//           +------------+
// 
   
   int   bRxIdle:1;
   int   bRxBuffering:1;
   int   bRxActive:1;
   struct AdpcmTAG {
      union {
         soundbuf SfHdr;
         rtp_hdr_t RtpHdr;
      } u;
      char AdpcmStateBuf[4];
      char Buf[ADPCM_PACKET_DATA_BYTES];
      char AdpcmStateBuf1[3];
   } GCC_PACKED Adpcm;
   gsm hGSM_Encode;
   
   int ADPCM_Bytes; // number of bytes of compressed ADPCM data in buffer
   
   struct adpcm_state TxADPCMStart;
   struct adpcm_state TxADPCM;
   u_int16 TxSequence;
   u_int16 RtpTxSequence;
   char GSM_Packet[SIZEOF_SOUNDBUF + (GSM_FRAME_LEN * 12) + 2];
   int GsmFramesAvail;

   void DecompressRTP();
   void DecompressGSM(gsm_byte *pGSM,int Len,gsm gsmh);
   void DecompressADPCM(char *p,int Len,struct adpcm_state *pState);
   void InsertInRTPQ(RtpData *pRtpData,int MaxNetQueueLen);
   int  DeleteTx(LinkedNode *pLink);
   void AddTx(LinkedNode *pLink);
   void ConsumeAudio(NetAudio *pAudio);
   void AdpcmTx();
   void GsmTx();
   void SendAudioPacket(void *Buf,int TxLen);
   void Add2CtoneTimer(int AudioBytes);
   NetAudio *AllocAudioBuf(int AudioSamples);
};

void EndPointCleanup(ConfClient *pCC);

#endif // __VOIP_H_
